#****************************************************************************
#**
#**  File     :  /lua/sim/buff.lua
#**
#**  Copyright  2008 Gas Powered Games, Inc.  All rights reserved.
#****************************************************************************

# The Unit's BuffTable for applied buffs looks like this:
#
# Unit.Buffs = {
#    Affects = {
#        <AffectType (Regen/MaxHealth/etc)> = {
#            BuffName = {
#                Count = i,
#                Add = X,
#                Mult = X,
#            }
#        }
#    }
#    BuffTable = {
#        <BuffType (LEVEL/CATEGORY)> = {
#            BuffName = {
#                Count = i,
#                Trash = trashbag,
#            }
#        }
#    }

#Function to apply a buff to a unit.
#This function is a fire-and-forget.  Apply this and it'll be applied over time if there is a duration.
function ApplyBuff(unit, buffName, instigator)
    if unit:IsDead() then 
        return 
    end
    #LOG ('buf.apply:',buffName)
    instigator = instigator or unit

    #buff = table of buff data
    local def = Buffs[buffName]
    #LOG('*BUFF: FullBuffTable: ', repr(Buffs))
    #LOG('*BUFF: BuffsTable: ', repr(def))
    if not def then
        error("*ERROR: Tried to add a buff that doesn\'t exist! Name: ".. buffName, 2)
        return
    end
    
    if def.EntityCategory then
        local cat = ParseEntityCategory(def.EntityCategory)
        if not EntityCategoryContains(cat, unit) then
            return
        end
    end
    
    if def.BuffCheckFunction then
        if not def:BuffCheckFunction(unit) then
            #LOG('return buff check')
            return
        end
    end
    
    local ubt = unit.Buffs.BuffTable
    
    if def.MinLevel and def.MinLevel > unit.VeteranLevel then return end
    
    if def.MaxLevel and def.MaxLevel < unit.VeteranLevel then return end
    
    
    if def.Stacks == 'REPLACE' and ubt[def.BuffType] then
        for key, bufftbl in unit.Buffs.BuffTable[def.BuffType] do
            RemoveBuff(unit, key, true)
        end
    end

    
    #If add this buff to the list of buffs the unit has becareful of stacking buffs.
    if not ubt[def.BuffType] then
        ubt[def.BuffType] = {}
    end
    
    if def.Stacks == 'IGNORE' and ubt[def.BuffType] and table.getsize(ubt[def.BuffType]) > 0 then
        #LOG('return Ignore')
        return
    end
    
    local data = ubt[def.BuffType][buffName]
    if not data then
        # This is a new buff (as opposed to an additional one being stacked)
        data = {
            Count = 1,
            Trash = TrashBag(),
            BuffName = buffName,
        }
        ubt[def.BuffType][buffName] = data
    else
        # This buff is already on the unit so stack another by incrementing the
        # counts. data.Count is how many times the buff has been applied
        data.Count = data.Count + 1
        
    end
    
    local uaffects = unit.Buffs.Affects
    if def.Affects then
        for k,v in def.Affects do
            # Don't save off 'instant' type affects like health and energy
            if k != 'Health' and k != 'Energy' then
                if not uaffects[k] then
                    uaffects[k] = {}
                end
                
                if not uaffects[k][buffName] then
                    # This is a new affect.
                    local affectdata = { 
                        BuffName = buffName, 
                        Count = 1, 
                    }
                    for buffkey, buffval in v do
                        affectdata[buffkey] = buffval
                    end
                    uaffects[k][buffName] = affectdata
                else
                    # This affect is already there, increment the count
                    uaffects[k][buffName].Count = uaffects[k][buffName].Count + 1
                end
            end
        end
    end
    
    #If the buff has a duration, then 
    if def.Duration and def.Duration > 0 then
        local thread = ForkThread(BuffWorkThread, unit, buffName, instigator) 
        unit.Trash:Add(thread)
        data.Trash:Add(thread)
    end
    
    PlayBuffEffect(unit, buffName, data.Trash)
    
    ubt[def.BuffType][buffName] = data

    if def.OnApplyBuff then
        def:OnApplyBuff(unit, instigator)
    end
  
    #LOG('*DEBUG: Applying buff :',buffName, ' to unit ',unit:GetUnitId())
    #LOG('Buff = ',repr(ubt[def.BuffType][buffName]))
    #LOG('Affects = ',repr(uaffects))
    BuffAffectUnit(unit, buffName, instigator, false)
end

#Function to do work on the buff.  Apply it over time and in pulses.
function BuffWorkThread(unit, buffName, instigator)
    
    local buffTable = Buffs[buffName]
    
    #Non-Pulsing Buff
    local totPulses = buffTable.DurationPulse
    
    if not totPulses then
        WaitSeconds(buffTable.Duration)
    else
        local pulse = 0
        local pulseTime = buffTable.Duration / totPulses
    
        while pulse <= totPulses and not unit:IsDead() do
    
            WaitSeconds(pulseTime)
            BuffAffectUnit(unit, buffName, instigator, false)
            pulse = pulse + 1
    
        end
    end

    RemoveBuff(unit, buffName)
end

#Function to affect the unit.  Everytime you want to affect a new part of unit, add it in here.
#afterRemove is a bool that defines if this buff is affecting after the removal of a buff.  
#We reaffect the unit to make sure that buff type is recalculated accurately without the buff that was on the unit.
#However, this doesn't work for stunned units because it's a fire-and-forget type buff, not a fire-and-keep-track-of type buff.
#This is because all timed buffs expire after their time is up
function BuffAffectUnit(unit, buffName, instigator, afterRemove)
    
    local buffDef = Buffs[buffName]
    
    local buffAffects = buffDef.Affects
    
    if buffDef.OnBuffAffect and not afterRemove then
        buffDef:OnBuffAffect(unit, instigator)
    end
    #LOG('Beginning to affect unit with buffs!')
    #LOG(repr(buffDef))
    for atype, vals in buffAffects do
    
        if atype == 'Health' then
        
            #Note: With health we don't actually look at the unit's table because it's an instant happening.  We don't want to overcalculate something as pliable as health.
            
            local health = unit:GetHealth()
            local val = ((buffAffects.Health.Add or 0) + health) * (buffAffects.Health.Mult or 1)
            local healthadj = val - health
            
            if healthadj < 0 then
                # fixme: DoTakeDamage shouldn't be called directly
                local data = {
                    Instigator = instigator,
                    Amount = -1 * healthadj,
                    Type = buffDef.DamageType or 'Spell',
                    Vector = VDiff(instigator:GetPosition(), unit:GetPosition()),
                }
                unit:DoTakeDamage(data)
            else
                unit:AdjustHealth(instigator, healthadj)
            
                #LOG('*BUFF: Unit ', repr(unit:GetEntityId()), ' buffed health to ', repr(val))
            end
        
        elseif atype == 'MaxHealth' then
        
            local unitbphealth = unit:GetBlueprint().Defense.MaxHealth or 1
            local val = BuffCalculate(unit, buffName, 'MaxHealth', unitbphealth)
        
            local oldmax = unit:GetMaxHealth()
            local oldcurrent = unit:GetHealth()
            val = math.ceil(val)
            unit:SetMaxHealth(val)
            
            unit:SetHealth(unit, val * oldcurrent/oldmax) 
            
            #if val > oldmax then
            #    unit:AdjustHealth(unit, val - oldmax)
            #else
            #    unit:SetHealth(unit, math.min(unit:GetHealth(), unit:GetMaxHealth())) 
            #end
            
            
            #LOG('*BUFF: Unit ', repr(unit:GetEntityId()), ' buffed max health to ', repr(val))
        #Values used when doing field repairs
        elseif atype == 'SelfRepair' then
            local val = BuffCalculate(unit, buffName, 'SelfRepair',unit:GetBlueprint().Defense.SelfRepairRate or 0)
            unit.UnitSelfRepairRate = val
            unit:ApplyRegen()
            
        #Values used when repairing a unit
        elseif atype == 'Repair' then
            local val = BuffCalculate(unit, buffName, 'Repair', 0)
            unit.UnitBeingRepairedRate = val
            unit:ApplyRegen()
        #Regen stores the values used for 'damage sponge' or original type regen
        elseif atype == 'Regen' or atype == 'RegenPercent' then
            local regen = unit:GetBlueprint().Defense.RegenRate or 0
            local val = BuffCalculate(unit, buffName, 'Regen', regen)
            local regenperc, bool, exists = BuffCalculate(unit, buffName, 'RegenPercent', unit:GetMaxHealth())
            if exists then  
                val = val +  regenperc
            end
            unit.UnitRegenRate = val
            unit:ApplyRegen()
            
--          Removed to fix the veterancy + enhancements overwriting each other

            #LOG('*BUFF: Unit ', repr(unit:GetEntityId()), ' buffed regen rate to ', repr(val))
--         elseif atype == 'RegenPercent' then
--         
--          local bpregn = unit:GetBlueprint().Defense.RegenRate or 0
--             local val = BuffCalculate(unit, buffName, 'Regen', bpregn)
--             LOG(val)
--          local val = val + BuffCalculate(unit, buffName, 'RegenPercent', unit:GetMaxHealth())
--             unit:SetRegenRate(val)
--             unit.Sync.RegenRate = val
--             LOG('*BUFF: Unit ', repr(unit:GetEntityId()), ' buffed regen rate to ', repr(val))
--            local val = false
                
--             if afterRemove then
--                 #Restore normal regen value plus buffs so I don't break stuff. Love, Robert
--                 local bpregn = unit:GetBlueprint().Defense.RegenRate or 0
--                 val = BuffCalculate(unit, nil, 'Regen', bpregn)
--             else
--                 #Buff this sucka
--                 val = BuffCalculate(unit, buffName, 'RegenPercent', unit:GetMaxHealth())
--             end
--             
--             unit:SetRegenRate(val)
--             unit.Sync.RegenRate = val
            
        elseif atype == 'ShieldHP' then
            
            local shield = unit:GetShield()
            if not shield then return end
            local val = shield:GetBaseHP() or 0
            val = BuffCalculate(unit, buffName, atype, val)
            val = math.ceil(val)
            shield:SetMaxHealth(val)
             
            #LOG('*BUFF: Unit ', repr(unit:GetEntityId()), ' buffed max shieldhealth to ', repr(val))
            
         elseif atype == 'ShieldRegen' then
            
            local shield = unit:GetShield()
            if not shield then return end
            local spec = shield:GetSpec()
            local valregen = spec.ShieldRegenRate
            #local valrecharge = spec.ShieldRechargeTime
            #local valrechargeenergy = spec.ShieldEnergyDrainRechargeTime
            valregen = BuffCalculate(unit, buffName, atype, valregen)
            
            shield:SetShieldRegenRate(valregen)
            #LOG('*BUFF: Unit ', repr(unit:GetEntityId()), ' buffed shieldregen to ', repr(valregen))
            
        elseif atype == 'Damage' then
            for i = 1, unit:GetWeaponCount() do
                local wep = unit:GetWeapon(i)
                local wepbp = wep:GetBlueprint()
                if wep.Label != 'DeathWeapon' and wep.Label != 'DeathImpact' then
                    if not (wepbp.WeaponCategory == 'Death' or vals.ByName and not vals.ByName[wepbp.Label] ) then
                    local wepbp = wep:GetBlueprint()
                        local val = BuffCalculate(unit, buffName, atype, wepbp.Damage)
                        if val >= ( math.abs(val) + 0.5 ) then
                            val = math.ceil(val)
                        else
                            val = math.floor(val)
                        end
                        wep.Damage = val
                        #LOG('*BUFF: Unit ', repr(unit:GetEntityId()), ' buffed damage to ', repr(val))
                        
                        if wepbp.NukeOuterRingDamage and wepbp.NukeInnerRingDamage then
                            unit.NukeOuterRingDamage = BuffCalculate(unit, buffName, atype, wepbp.NukeOuterRingDamage)
                            unit.NukeInnerRingDamage = BuffCalculate(unit, buffName, atype, wepbp.Damage)
                        end
                    end
                end
            end
        
        elseif atype == 'DamageRadius' then
            for i = 1, unit:GetWeaponCount() do
                local wep = unit:GetWeapon(i)
                local wepbp = wep:GetBlueprint()
                if not (wepbp.WeaponCategory == 'Death' or vals.ByName and not vals.ByName[wepbp.Label] ) then
                    local val = BuffCalculate(unit, buffName, atype, wepbp.DamageRadius)
                    wep.DamageRadius = val
                    #LOG('*BUFF: Unit ', repr(unit:GetEntityId()), ' buffed AoE to ', repr(val))
                    
                    if wepbp.NukeOuterRingRadius and wepbp.NukeInnerRingRadius then
                        unit.NukeOuterRingRadius = (BuffCalculate(unit, buffName, atype, wepbp.NukeOuterRingRadius)+wepbp.NukeInnerRingRadius)/2
                        unit.NukeInnerRingRadius = (BuffCalculate(unit, buffName, atype, wepbp.NukeInnerRingRadius)+wepbp.NukeInnerRingRadius)/2  
                    end
                end
            end

        elseif atype == 'MaxRadius' then
            for i = 1, unit:GetWeaponCount() do
                local wep = unit:GetWeapon(i)
                local wepbp = wep:GetBlueprint()
                if not (wepbp.WeaponCategory == 'Death' or vals.ByName and not vals.ByName[wepbp.Label] ) then
                    local val = BuffCalculate(unit, buffName, atype, wepbp.MaxRadius)
                    #LOG(wepbp.Label .. ' newRange:' .. val)
                    wep:ChangeMaxRadius(val)
                    #Used to increase lifetime of projectile so it doesn't disappear before max range.
                    wep.rangeMod = val / wepbp.MaxRadius
                end
                
                #LOG('*BUFF: Unit ', repr(unit:GetEntityId()), ' buffed max radius to ', repr(val))
            end
        

        elseif atype == 'MoveMult' then
            local UnitType = unit:GetBlueprint().Categories
            #LOG(repr(UnitType))
            if not table.find(UnitType,'AIR') then
                local val = BuffCalculate(unit, buffName, 'MoveMult', 1)
                unit:SetSpeedMult(val)
                unit:SetAccMult(val)
                unit:SetTurnMult(val)
                #LOG('*BUFF: Unit ', repr(unit:GetEntityId()), ' buffed speed/accel/turn mult to ', repr(val))
            end
        elseif atype == 'Stun' and not afterRemove then
            
            unit:SetStunned(buffDef.Duration or 1, instigator)
            
            #LOG('*BUFF: Unit ', repr(unit:GetEntityId()), ' buffed stunned for ', repr(buffDef.Duration or 1))
            
            if unit.Anims then
                for k, manip in unit.Anims do
                    manip:SetRate(0)
                end
            end
                    
        elseif atype == 'WeaponsEnable' then
            
            for i = 1, unit:GetWeaponCount() do
                local wep = unit:GetWeapon(i)
                if not (wepbp.WeaponCategory == 'Death' or vals.ByName and not vals.ByName[wepbp.Label] ) then               
                    local val, bool = BuffCalculate(unit, buffName, 'WeaponsEnable', 0, true)
                    wep:SetWeaponEnabled(bool)
                end
            end
            
        elseif atype == 'VisionRadius' then
            local val = BuffCalculate(unit, buffName, 'VisionRadius', unit:GetBlueprint().Intel.VisionRadius or 0)
            unit:SetIntelRadius('Vision', val)
            #LOG('*BUFF: Unit ', repr(unit:GetEntityId()), ' buffed Vison to ', repr(val))

        elseif atype == 'RadarRadius' then
            local val = BuffCalculate(unit, buffName, 'RadarRadius', unit:GetBlueprint().Intel.RadarRadius or 0)
            
            if val <= 0 then
                unit:DisableIntel('Radar')
                return
            end
        
            
            if not unit:IsIntelEnabled('Radar') then
                unit:InitIntel(unit:GetArmy(),'Radar', val)
                unit:EnableIntel('Radar')
            else
                unit:SetIntelRadius('Radar', val)
                unit:EnableIntel('Radar')
            end
            
            
        elseif atype == 'OmniRadius' then
            local val = BuffCalculate(unit, buffName, 'OmniRadius', unit:GetBlueprint().Intel.OmniRadius or 0)
            
            if val <= 0 then
                unit:DisableIntel('Omni')
                return
            end            
            
            
             if not unit:IsIntelEnabled('Omni') then
                unit:InitIntel(unit:GetArmy(),'Omni', val)
                unit:EnableIntel('Omni')
            else
                unit:SetIntelRadius('Omni', val)
                unit:EnableIntel('Omni')
            end
                        
        elseif atype == 'BuildRate' then
            local val = BuffCalculate(unit, buffName, 'BuildRate', unit:GetBlueprint().Economy.BuildRate or 1)
            unit:SetBuildRate( val )
            unit.Sync.BuildRate = val
            
        elseif atype == 'EnergyProductionBuf' then
            local val = BuffCalculate(unit, buffName, 'EnergyProductionBuf', unit:GetBlueprint().Economy.ProductionPerSecondEnergy or 0)
            unit.EnergyProdMod = val
            unit:UpdateProductionValues()

        elseif atype == 'MassProductionBuf' then
            local val = BuffCalculate(unit, buffName, 'MassProductionBuf', unit:GetBlueprint().Economy.ProductionPerSecondMass or 0)
            unit.MassProdMod = val
            unit:UpdateProductionValues()
            
            
        #### ADJACENCY BELOW ####
        elseif atype == 'EnergyActive' then
            local val = BuffCalculate(unit, buffName, 'EnergyActive', 1)
            unit.EnergyBuildAdjMod = val
            unit:UpdateConsumptionValues()
            #LOG('*BUFF: EnergyActive = ' ..  val)
            
        elseif atype == 'MassActive' then
            local val = BuffCalculate(unit, buffName, 'MassActive', 1)
            unit.MassBuildAdjMod = val
            unit:UpdateConsumptionValues()
            #LOG('*BUFF: MassActive = ' ..  val)
            
        elseif atype == 'EnergyMaintenance' then
            local val = BuffCalculate(unit, buffName, 'EnergyMaintenance', 1)
            unit.EnergyMaintAdjMod = val
            unit:UpdateConsumptionValues()
            #LOG('*BUFF: EnergyMaintenance = ' ..  val)
            
        elseif atype == 'MassMaintenance' then
            local val = BuffCalculate(unit, buffName, 'MassMaintenance', 1)
            unit.MassMaintAdjMod = val
            unit:UpdateConsumptionValues()
            #LOG('*BUFF: MassMaintenance = ' ..  val)
            
        elseif atype == 'EnergyProduction' then
            local val = BuffCalculate(unit, buffName, 'EnergyProduction', 1)
            unit.EnergyProdAdjMod = val
            unit:UpdateProductionValues()
            #LOG('*BUFF: EnergyProduction = ' .. val)

        elseif atype == 'MassProduction' then
            local val = BuffCalculate(unit, buffName, 'MassProduction', 1)
            unit.MassProdAdjMod = val
            unit:UpdateProductionValues()
            #LOG('*BUFF: MassProduction = ' .. val)
            
        elseif atype == 'EnergyWeapon' then
            local val = BuffCalculate(unit, buffName, 'EnergyWeapon', 1)
            for i = 1, unit:GetWeaponCount() do
                local wep = unit:GetWeapon(i)
                if wep:WeaponUsesEnergy() then
                    wep.AdjEnergyMod = val
                end
            end
            #LOG('*BUFF: EnergyWeapon = ' ..  val)
            
         elseif atype == 'RateOfFire' then
            for i = 1, unit:GetWeaponCount() do
                 local wep = unit:GetWeapon(i)
                 local wepbp = wep:GetBlueprint()
                 local weprof = wepbp.RateOfFire

                 # Set new rate of fire based on blueprint rate of fire.=
                 if not (wepbp.WeaponCategory == 'Death' or vals.ByName and not vals.ByName[wepbp.Label] ) then
                     local val = BuffCalculate(unit, buffName, 'RateOfFire', 1)

                     #local delay = 1 / wepbp.RateOfFire
                     wep.adjRoF= val
                     wep:ChangeRateOfFire(wep.bufRoF/wep.adjRoF)
                 end
                 #wep:ChangeRateOfFire( 1 / ( val * delay ) )
                #LOG(string.format('*BUFF: RateOfFire = %f (val:%f)' , 1 / ( val * delay ),val))
            end
        elseif atype == 'RateOfFireBuf' then  --New function inserted to handle ROF buffs by name 
            for i = 1, unit:GetWeaponCount() do
                local wep = unit:GetWeapon(i)
                local wepbp = wep:GetBlueprint()
                if not (wepbp.WeaponCategory == 'Death' or vals.ByName and not vals.ByName[wepbp.Label] ) then
                    local val = BuffCalculate(unit, buffName, atype, wepbp.RateOfFire)
                    #LOG(wepbp.Label .. ' newRoF:' .. val)
                    wep.bufRoF = val
                    wep:ChangeRateOfFire(wep.bufRoF/wep.adjRoF)
                end
            end    
        elseif atype == 'DPS' then
            local val = BuffCalculate(unit, buffName, 'DPS', unit:GetBlueprint().dps or 0)
            unit.DPS = val


#   CLOAKING is a can of worms.  Revisit later.
#        elseif atype == 'Cloak' then
#            
#            local val, bool = BuffCalculate(unit, buffName, 'Cloak', 0)
#            
#            if unit:IsIntelEnabled('Cloak') then
#
#                if bool then
#                    unit:InitIntel(unit:GetArmy(), 'Cloak')
#                    unit:SetRadius('Cloak')
#                    unit:EnableIntel('Cloak')
#            
#                elseif not bool then
#                    unit:DisableIntel('Cloak')
#                end
#            
#            end
           
        elseif atype != 'Stun' then
            WARN("*WARNING: Tried to apply a buff with an unknown affect type of " .. atype .. " for buff " .. buffName)
        end
    end
end

#BuffCalculate(unit, buffName, 'RateOfFire', wepbp.RateOfFire)

#Calculates the buff from all the buffs of the same time the unit has.
function BuffCalculate(unit, buffName, affectType, initialVal, initialBool)
    
    #Add all the 
    local adds = 0
    local mults = 1.0
    local exists = false
    local divs = 1.0
    local bool = initialBool or false
    
    local highestCeil = false
    local lowestFloor = false
    
    
    if not unit.Buffs.Affects[affectType] then
        #LOG(affectType .. ' is missing!')
        #LOG(repr(unit.Buffs.Affects))
        return initialVal, bool, exists 
    end
    
    for k, v in unit.Buffs.Affects[affectType] do
        exists = true
        
        if v.Add and v.Add != 0 then
            adds = adds + (v.Add * v.Count)
        end
     
        if v.Mult then
            for i=1,v.Count do
                if v.Mult >= 1 then
                    mults = mults * v.Mult
                else
                    divs = divs * v.Mult
                end
            end
        end
        
        if not v.Bool then
            bool = false
        else
            bool = true
        end
        
        if v.Ceil and (not highestCeil or highestCeil < v.Ceil) then
            highestCeil = v.Ceil
        end
        
        if v.Floor and (not lowestFloor or lowestFloor > v.Floor) then
            lowestFloor = v.Floor
        end
    end
    
--     if not domult then
--      mults = 0
--      divs = 0
--     end
    
    #Adds are calculated first, then the mults.  May want to expand that later.
    local returnVal = (initialVal + adds) * mults * divs
    if lowestFloor and returnVal < lowestFloor then returnVal = lowestFloor end
    
    if highestCeil and returnVal > highestCeil then returnVal = highestCeil end 
    
    
    #LOG('*BUFFCALC: Type:' .. affectType ..' initialVal:' ..  initialVal .. ' adds:' .. adds .. ' mults:' .. mults .. ' returnVal:' .. returnVal)
    return returnVal, bool, exists
end



#Removes buffs
function RemoveBuff(unit, buffName, removeAllCounts, instigator)
    
    local def = Buffs[buffName]
    #LOG('BUFF to remove :: ',repr (def))
    local unitBuff = unit.Buffs.BuffTable[def.BuffType][buffName]
    
    for atype,_ in def.Affects do
        local list = unit.Buffs.Affects[atype]
        if list and list[buffName] then
            # If we're removing all buffs of this name, only remove as 
            # many affects as there are buffs since other buffs may have
            # added these same affects.
            if removeAllCounts then
                list[buffName].Count = list[buffName].Count - unitBuff.Count
            else
                list[buffName].Count = list[buffName].Count - 1
            end
            
            if list[buffName].Count <= 0 then
                list[buffName] = nil
            end
        end
    end
    
    
    if not unitBuff.Count then
        local stg = "*WARNING: BUFF: unitBuff.Count is nil.  Unit: "..unit:GetUnitId().." Buff Name: ".. buffName.." Unit BuffTable: ", repr(unitBuff)
        error(stg, 2)
    else
        unitBuff.Count = unitBuff.Count - 1
    end

    if removeAllCounts or unitBuff.Count <= 0 then
        # unit:PlayEffect('RemoveBuff', buffName)
        unitBuff.Trash:Destroy()
        unit.Buffs.BuffTable[def.BuffType][buffName] = nil
    end

    if def.OnBuffRemove then
        def:OnBuffRemove(unit, instigator)
    end

    # FIXME: This doesn't work because the magic sync table doesn't detect
    # the change. Need to give all child tables magic meta tables too.
    if def.Icon then
        # If the user layer was displaying an icon, remove it from the sync table
        local newTable = unit.Sync.Buffs
        table.removeByValue(newTable,buffName)
        unit.Sync.Buffs = table.copy(newTable)
    end

    BuffAffectUnit(unit, buffName, unit, true)
    
    #LOG('*BUFF: Removed ', buffName)
end

function HasBuff(unit, buffName)
    local def = Buffs[buffName]
    if not def then
        return false
    end
    local bonu = unit.Buffs.BuffTable[def.BuffType][buffName]
    if bonu then
        return true
    end
    return false
end

function PlayBuffEffect(unit, buffName, trsh)
    
    local def = Buffs[buffName]
    if not def.Effects then 
        return 
    end
    
    for k, fx in def.Effects do
        local bufffx = CreateAttachedEmitter(unit, 0, unit:GetArmy(), fx)
        if def.EffectsScale then
            bufffx:ScaleEmitter(def.EffectsScale)
        end
        trsh:Add(bufffx)
        unit.TrashOnKilled:Add(bufffx)
    end
end

#
# DEBUG FUNCTIONS
# 
_G.PrintBuffs = function()
    local selection = DebugGetSelection()
    for k,unit in selection do
        if unit.Buffs then
            LOG('Buffs = ', repr(unit.Buffs))
        end
    end
end

_G.UpdateRegen = function(value)
    local selection = DebugGetSelection()
    for k,unit in selection do
        unit.UnitRegenRate = value
        unit:ApplyRegen()
    end
end

_G.PrintStats = function()
    local selection = DebugGetSelection()
    local csv = 'UnitName, BlueprintId, MaxHealth, EnergyCost, Mass Cost, Build Time, BuildRate, '
    csv = csv .. 'Max Accel, Max Speed, Max Brake, Turn Rate, Turn Radius, Weapon, Damage, DamageRadius, FiringRandomness, FiringTolerance, MaxRadius, RateOfFire'
    for k,unit in selection do
        csv = csv .. '\n'
        bp = unit:GetBlueprint()
        csv = csv .. string.format(bp.General.UnitName  or '')
        csv = csv .. ',' .. string.format(bp.BlueprintId or '')
        csv = csv .. ',' .. string.format(bp.Defense.MaxHealth or '')
        csv = csv .. ',' .. string.format(bp.Economy.BuildCostEnergy or '')
        csv = csv .. ',' .. string.format(bp.Economy.BuildCostMass or '')
        csv = csv .. ',' .. string.format(bp.Economy.BuildTime or '')
        csv = csv .. ',' .. string.format(bp.Economy.BuildRate or '')
        csv = csv .. ',' .. string.format(bp.Physics.MaxAcceleration or '')
        csv = csv .. ',' .. string.format(bp.Physics.MaxSpeed or '')
        csv = csv .. ',' .. string.format(bp.Physics.MaxBrake or '')
        csv = csv .. ',' .. string.format(bp.Physics.TurnRate or '')
        csv = csv .. ',' .. string.format(bp.Physics.TurnRadius or '')
        for k2, wep in bp.Weapon do
            csv = csv .. ',' ..     string.format(wep.Label or '')
            csv = csv .. ',' ..     string.format(wep.Damage or '')
            csv = csv .. ',' ..     string.format(wep.DamageRadius or '')
            csv = csv .. ',' ..     string.format(wep.FiringRandomness or '')
            csv = csv .. ',' ..     string.format(wep.FiringTolerance or '')
            csv = csv .. ',' ..     string.format(wep.MaxRadius or '')
            csv = csv .. ',' ..     string.format(wep.RateOfFire or '')
        end
    end
    LOG(repr(csv))
end

_G.PrintXML = function()
    local csv = ""
    local selection = DebugGetSelection()
    for k,unit in selection do
        bp = unit:GetBlueprint()
        LOG(type(bp))
        csv = csv .. auto_xml(bp.BlueprintId,bp.General.UnitName or 'Unit',bp,'')
    end
    LOG(csv)
end

all_blueprints = {}
current_filename = nil

#
# Strip localization crud from a string
#
local function LOC(s)
    if string.sub(s, 1, 4)=='<LOC' then
        local i = string.find(s,">")
        if i then
            s = string.sub(s, i+1)
        end
    end
    return s
end


#
# Escape problematic XML entities in a string
#
local function ESCAPE(s)
    s = string.gsub(s, '&', '&amp;')
    s = string.gsub(s, '<', '&lt;')
    s = string.gsub(s, '>', '&gt;')
    return s
end


#
# Define 'UnitBlueprint' func, which will get called with the blueprint data
#
function UnitBlueprint(spec)
    spec._xml_tag = "UnitBlueprint"
    spec._xml_attrs = { id=string.gsub(string.lower(current_filename), "^.*/([^/]+)_[a-z]+%.bp$", "%1") }

    table.insert(all_blueprints,spec)
end

function Sound(t)
    t._xml_tag = 'Sound'
    return t
end

function RPCSound(t)
    t._xml_tag = 'RPCSound'
    return t
end

# Dump them in XML
#
# stuff = { a,b,c } becomes <stuff><li>a</li><li>b</li><li>c</li></stuff>
#
# { x=this, y=that } becomes <x>this</x><y>that</y>
#
function xml_tags(defaulttag, name, value)
    local tag = value._xml_tag or defaulttag
    local close = '</' .. tag .. '>'

    local open = '<' .. tag
    if name and name!=tag then
        open = open .. ' name="' .. name .. '"'
    end
    if value._xml_attrs then
        for k,v in value._xml_attrs do
            open = open .. (' ' .. k .. '="' .. v .. '"')
        end
    end
    open = open .. '>'

    return open,close
end


function auto_xml(tag,name,value,indent)
    local csv = ""
    local open,close = xml_tags(tag, name, value)

    if type(value)=='table' then
        local subindent = indent..'  '
        csv = csv .. open .. '\n'
        if table.getn(value)>0 then
            # numerically indexed table
            for i,v in ipairs(value) do
                csv = csv ..subindent
                csv = csv ..auto_xml('li',nil,v,subindent)
            end
        else
            # key=value table
            for k,v in value do
                --if string.sub(k,1,1)!='_' then
                    csv = csv ..subindent
                    csv = csv .. auto_xml(k,k,v,subindent)
                --end
            end
        end
        csv = csv ..indent..close.. '\n'
    else
        if type(value)=='string' then
            value = ESCAPE(LOC(value))
        else
            value = tostring(value)
        end
        csv = csv ..open
        csv = csv ..value
        csv = csv ..close .. '\n'
    end
    return csv
end
